﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Core;

namespace GradientBoostedTrees {
  public class SquaredErrorLoss : ILossFunction {
    public double GetLoss(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      double s = 0;
      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        double res = targetEnum.Current - predEnum.Current;
        s += weightEnum.Current * res * res;
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have differing lengths");

      return s;
    }

    public IEnumerable<double> GetLossGradient(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        yield return weightEnum.Current * 2.0 * (targetEnum.Current - predEnum.Current);
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have differing lengths");
    }

    public LineSearchFunc GetLineSearchFunc(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetArr = target.ToArray();
      var predArr = pred.ToArray();
      var weightArr = weight.ToArray();
      if (targetArr.Length != predArr.Length || predArr.Length != weightArr.Length)
        throw new ArgumentException("target, pred and weight have differing lengths");

      // line search for 
      LineSearchFunc lineSearch = (idx, startIdx, endIdx) => {
        double s = 0.0;
        int n = 0;
        for (int i = startIdx; i <= endIdx; i++) {
          int row = idx[i];
          s += (targetArr[row] - predArr[row]);
          n++;
        }
        return s / n;
      };
      return lineSearch;

    }

    public override string ToString() {
      return "Squared error loss";
    }
  }
}
